// *************************************************************************
//    HashIndex.h
// *************************************************************************

#ifndef _HashIndex_H_
#define _HashIndex_H_

#include "GenericIndex.h"
#include "TupleStack.h"

// *************************************************************************

#define UnmatchedHashValue ((HashVal) -3)

// *************************************************************************

class Cor_HashIndex : public Cor_GenericIndex {

  friend class Cor_HashIndexSet;

  protected:
    int hash_size;                  // number of buckets in the hash index
    int limit;                      // Grow hash table if there are more
                                    // than this many tuples

    double incr_ratio, max_occupancy;

    int hash_count;                 // number of Tuples in hash_table,
                                    // but not in var_chain

    Cor_TupleStack **hash_table;
    Cor_TupleStack  *var_chain;     // tuples whose hash value is VarHashValue

    // The next field implements a scheme for only scanning the
    // non-empty buckets of the relation, so that empty buckets are
    // never considered during a file scan.

    Cor_IntStack   *non_mt_bckts;

  public: 

    Cor_HashIndex( Cor_StorageRelation *rel, Cor_BitVector *aset,
        int buckets = -1, double incr_ratio1 = -1.0,
        double max_occupancy1 = -1.0);

    Cor_HashIndex( Cor_StorageRelation *rel, Cor_ArgList *pat,
        int n_var, Cor_ArgList *v_nm,
        int buckets = -1, double incr_ratio1 = -1.0,
        double max_occupancy1 = -1.0);

    virtual ~Cor_HashIndex();

    void initialize_iterator(Cor_TupleIterator&);

    Cor_Tuple *get_next_tuple(Cor_TupleIterator& it);

    Cor_BindEnv *get_next(Cor_TupleIterator&);

    HashVal hash(Cor_ArgList *args, Cor_BindEnv *bindenv);
    int insert_tuple(Cor_Tuple *tuple);

    int is_subsumed_bkt(Cor_Tuple *tuple,
            HashVal hash_val, Cor_TupleStack *ts);
    virtual int is_subsumed(Cor_Tuple *tuple);

    virtual int index_type()    {return COR_HASH_INDEX;}

    int  insert_entry(HashVal h_val, Cor_Tuple *tuple);
    void rehash_bigger();

    int  delete_tuple(Cor_Tuple *tuple);

    void move_all_entries(Cor_GenericIndex *ind);
    void delete_all_tuples(int force_del=0);
    void reset_index();    

    // For debugging
    void show_all_tuples(ostream& out);
};

// *************************************************************************

#define DONT_USE_STATIC_ENV   1
#define MAX_VARS 64
#define VAR_INCR 8

// *************************************************************************

#endif _HashIndex_H_
