// *************************************************************************
//    PersistentRelation.h
// *************************************************************************

#ifdef  WITH_PERSISTENCE
#ifndef _PersistentRelation_H_
#define _PersistentRelation_H_

#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"
#include "RelationMisc.h"
#include "Relation.h"
#include "RelationList.h"
#include "PersistentIndex.h"
#include "PersistentIndexSet.h"
#include "Debug.h"


#define Scanning_Persistent 1

// *************************************************************************

class Cor_PersistentIndexSet;

class Cor_PersistentRel : public Cor_StorageRelation {
    friend class Cor_PersistentIndex;
    friend class Cor_PersistentIndexSet;
private:
    OID oid;                     // OID of FileDescStruct of the relation
    Cor_FileDescStruct fd;       // File descriptor

    LOADDESC *lddesc;            // bulk load descriptor

    Cor_PersistentIndexSet index_set;

    // Need representation of field types

    void insertObject(Cor_ArgList &arglist);
    Cor_Tuple *fileScanGetNext(Cor_TupleIterator &, int delete_flag = 0);
    Cor_Tuple *indexScanGetNext(Cor_TupleIterator&, int, int delete_flag = 0);

    // Returns position where an index entry can be made
    //int canAddIndex() { return fd.canAddIndex() ;} ;
    Cor_Tuple *checkTuple(USERDESC *, Cor_TupleIterator&);

public:
    Cor_PersistentRel(Cor_CArg& name, int arity, int& error);
    Cor_PersistentRel(Cor_CArg& name, int arity, int& error, OID*);
    Cor_PersistentRel(Cor_CArg& name, int arity,
        int& error, Cor_FileDescStruct *);

    Cor_FileDescStruct file_desc() { return(fd); }

    virtual void release(Cor_TupleIterator& iterator);
    virtual void print_name(ostream *file);
    virtual void print_facts(ostream *file,
            Cor_ArgList *arglist = NULL) const;

    virtual void freeMark(Cor_RMark *);
    virtual void free_allMarks();

    virtual Cor_BindEnv *get_next(Cor_TupleIterator&);
    virtual Cor_Tuple *get_next_tuple(Cor_TupleIterator&) ;
    virtual int insert_raw(Cor_Tuple *tuple);

    virtual int insert_tuple(Cor_Tuple *tuple);
    // maybe this can use the ::insert_new defined in generic-rel ????
    // virtual int insert_new(ArgList& args, BindEnv *env,
    //    BindEnv *dont_rename_env = NULL, Tuple *parent_goal = NULL,
    //    int dont_rename = 0); 

    virtual Cor_GenericIndexSet *get_index_set() { return &index_set; }
    virtual Cor_GenericIndex *create_index_object(Cor_BitVector *bv) {
        return new Cor_PersistentIndex(this, bv);
    }

    virtual Cor_GenericIndex *create_index_object(Cor_ArgList *,
            int, Cor_ArgList *) {
        return NULL;
    }

    virtual int tuple_delete(Cor_Tuple *tuple);

      // Empty tuples out of the relation
    virtual void empty_relation(int delete_tuples = 0); 
    void initialize_iterator( Cor_TupleIterator& );
    virtual int is_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark= NULL,
        Cor_RMark *endmark = NULL);

    int start_bulk_load();
    int end_bulk_load();
    void flush_to_disk();
};

// *************************************************************************

#endif // _PersistentRelation_H_
#endif // WITH_PERSISTENCE
