#include "RdbQuery.h"
#include "Log.h"


RdbQuery::RdbQuery(const RdbJoin& newJoin, const BitVector& newAdorn )
: RdbCompilableDML(newJoin.Database()),
  join(newJoin),
  adorn(newAdorn)
{  
    arity = adorn.len();
    assert(join.Arity() >= arity );
}


String RdbQuery::CreateStmtWithoutConditions()
{
    ostrstream s;
    s << "select ";
    int rels = join.RelCount();
    for( int i = 0 ; i < arity ; i++ ) {
	if( i > 0 ) s << ", ";
	s << VarQualifiedColName(i);
    }
    s << " from ";
    for( i = 0 ; i < rels ; i++ ) {
	if( i > 0 ) s << ", ";
	s << RelName(i) << " R" << i;
    }
    int joins = join.JoinCount();
    for( int j = 0 ; j < joins ; j++ ) { // take care of joins
	if( j == 0 ) s << " where ";
	else         s << " and ";
	s << join.VarJoinString(j);
    }
    return String(s);
}


void RdbQuery::CreateStmt(const String varPlaceHolder,
			  const int addSequence)
{
    ostrstream s;
    s << CreateStmtWithoutConditions();
    int needAnd = join.JoinCount() > 0;
    int v;
    int setBit;
    for(setBit = adorn.next(-1), v = 0;
	setBit >= 0 ;
	setBit = adorn.next(setBit), v++ ) {
	if( needAnd ) s << " and ";
	else          s << " where ";
	s << VarQualifiedColName(setBit) << " = " << varPlaceHolder;
	if( addSequence ) s << v;
	needAnd = TRUE;
    }
    statement = s;
    RDBLOG(database.DBLog(), "Select stmt: "<<statement<<'\n');
}


void RdbQuery::CreateStmt(const RdbItemList& bindings)
{
    ostrstream s;
    s << CreateStmtWithoutConditions();
    int needAnd = join.JoinCount() > 0;
    int i;
    int setBit;
    for(setBit = adorn.next(-1), i = 0;
	setBit >= 0 ;
	setBit = adorn.next(setBit), i++ ) {
	if( needAnd ) s << " and ";
	else          s << " where ";
	s << VarQualifiedColName(setBit);
	switch( bindings[i].Type() ) {
	  case RdbItem::RDB_STRING:
	    s << " = " << '\'' << bindings[i] << '\'';
	    break;
	  case RdbItem::RDB_INT:
	  case RdbItem::RDB_REAL:
	    s << " = " << bindings[i];
	    break;
	  case RdbItem::RDB_NULL:
	    s << " is null";
	    break;
	  default:
	    assert(0);
	}
	needAnd = TRUE;
    }
    statement = s;
    RDBLOG(database.DBLog(), "Select stmt: "<<statement<<'\n');
}


RdbError::ErrorCode RdbQuery::Open(RdbCursor& cursor)
{
    database.SetFocus();
    return VOpen(cursor);
}


RdbError::ErrorCode RdbQuery::Close(RdbCursor& cursor)
{
    database.SetFocus();
    return VClose(cursor);
}


RdbError::ErrorCode RdbQuery::GetNext(RdbCursor& cursor, RdbItemList& result)
{
    database.SetFocus();
    return VGetNext(cursor, result);
}

