#ifndef _LOGHANDLER_H_
#define _LOGHANDLER_H_

// LogHandler:
// 
// This class is used for quick and easy conditional output statements,
// like debugging statements.
// 
// Usage is normally through a macro, see Log.h.


#include <iostream.h>
#include <iomanip.h>


class LogHandler
{
  public:

    LogHandler(ostream* _pos=NULL, // stream to place output, NULL = disable
	       int _dispLoc=0,	   // boolean: true display file & line of call
	       int _delete_pos=0   // delete _pos? (in destructor or SetStream)
	       );

    ~LogHandler();

    void SetStream(ostream* _pos, int _delete_pos);
				// set the output stream
				// delete pos in destructor or SetStream?

    void Disable();		// turn off Log

    void Enable();		// turn on Log

    ostream& OutStream();	// return the output stream, valid only if on!

    int IsEnabled() const;	// should Log be printed?

    int ShowLocation() const;	// should the location be printed?

  private:

    ostream* pos;		// the stream to print on

    int enabled :1;		// the stream should be printed on 

    int dispLoc :1;		// the location of the call should be printed

    int delete_pos :1;		// delete the stream in destructor

    LogHandler(const LogHandler&); // n/a
    LogHandler& operator=(const LogHandler&); // n/a
};


#include "LogHandler.i"


#define LOG_MSG(LH,M) \
    if( (LH).IsEnabled() ) { \
        if( (LH).ShowLocation() ) \
	    (LH).OutStream() << '['<<__FILE__<<setw(5)<<__LINE__<<"]  "; \
	(LH).OutStream() << M; \
    } 


#endif // _LOGHANDLER_H_
