// *************************************************************************
//    ArgList.C
// *************************************************************************

#include <stdio.h>

#include "ArgList.h"

#include "BitVector.h"
#include "Globals.h"
#include "Stack.h"
#include "Debug.h"

// *************************************************************************
Cor_ArgList::Cor_ArgList()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    numArgs = 0;
    args = NULL;
}

// *************************************************************************
Cor_ArgList::Cor_ArgList(const Cor_ArgList& arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    numArgs = arglist.numArgs;
    args = new Cor_CArg[numArgs];

    for (int i = 0; i < numArgs; i++) 
        args[i] = arglist.args[i];
}

// *************************************************************************
Cor_ArgList::Cor_ArgList(const Cor_CArg& arg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    numArgs = 1;
    args = new Cor_CArg[numArgs];
    args[0] = arg;
}

// *************************************************************************
// Array already supplied
Cor_ArgList::Cor_ArgList(Cor_CArg *_args, int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    args = _args;
    numArgs = size;
}

// *************************************************************************
void Cor_ArgList::operator=(const Cor_ArgList& arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] args;

    numArgs = arglist.numArgs;
    args = new Cor_CArg[numArgs];

    for (int i = 0; i < numArgs; i++) 
        args[i] = arglist.args[i];
}

// *************************************************************************
Cor_ArgList::Cor_ArgList(int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    numArgs = size;
    args = new Cor_CArg[size];
}

// *************************************************************************
Cor_ArgList::~Cor_ArgList()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] args;
}

// *************************************************************************
void Cor_ArgList::set_count(int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (size == numArgs)
        return;

    if (size <= 0) {
        numArgs = 0;
        delete [] args;
        args = NULL;
        return;
    }

    delete [] args;
    numArgs = size;
    args = new Cor_CArg[size];
}
 
// *************************************************************************
void Cor_ArgList::print_dump(Cor_BindEnv *env, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //print_qv(env, cout_strm);
    print_qv(env, file);
}

// *************************************************************************
ostream& operator<< (ostream& out, const Cor_ArgList& al)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Note: Environment is assumed to be NULL for stream operator

    if (al.numArgs)
        out << al.args[0];

    for(int j=1; j < al.numArgs; j++) {
        out << ",";
        out << al.args[j];
    } 

    return(out);
}
 
// *************************************************************************
// quote variables (for Explain)
void Cor_ArgList::print_qv(Cor_BindEnv *env, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (numArgs)
        args[0].print_qv(env, file);

    for(int j=1; j < numArgs; j++) {
        *file << ",";
        args[j].print_qv(env, file);
    } 
}

void Cor_ArgList::print(Cor_BindEnv *env, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (numArgs)
        args[0].print(env, file);

    for(int j=1; j < numArgs; j++) {
        *file << ",";
        args[j].print(env, file);
    } 
}

void Cor_ArgList::literal_print(ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (numArgs)
        args[0].literal_print(file);

    for(int j=1; j < numArgs; j++) {
        *file << ",";
        args[j].literal_print(file);
    } 
}

// *************************************************************************
void Cor_ArgList::print(Cor_BindEnv *env, ostream *file, char *) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    print(env, file);
}

// *************************************************************************
void Cor_ArgList::sprint(char *str, int *pos, Cor_BindEnv *env) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int bFirst_arg = YES;
    FOR_EACH_ARG(arg, *this) {
        if (bFirst_arg == NO) {
            sprintf(str + *pos, ",");
            (*pos)++;
        } else
            bFirst_arg = NO;
        arg->sprint(str, pos, env);
    } END_EACH_ARG
}

// *************************************************************************
int Cor_ArgList::isConstant() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    FOR_EACH_ARG(arg, *this) {
        if (! arg->isConstant())
            return(0); 
    } END_EACH_ARG

    return (1);
}

// *************************************************************************
// Returns true if the arglist has grouping args
int Cor_ArgList::has_grouping_args()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    for (i=0; i < count(); i++)
        if ( args[i].is_grouping() )
            return(1);

    return(0);
}

// *************************************************************************
Cor_IntStack *Cor_ArgList::var_list(Cor_IntStack *var_stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!var_stk)
        var_stk = new Cor_IntStack();

    for (int j=0; j < numArgs; j++)
        args[j].var_list(var_stk);

    return(var_stk);
}

// *************************************************************************
Variable Cor_ArgList::max_vars() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int      j;
    Variable max_var_num = 0, tmp_max_var;

    for (j=0; j < numArgs; j++) {
        if ( (tmp_max_var = args[j].max_vars()) > max_var_num)
            max_var_num = tmp_max_var;
    }

    return(max_var_num);
}

// *************************************************************************
//  Computes the hash value of the set of arguments specified by
//  bitvector bv.
// *************************************************************************
HashVal Cor_ArgList::hash(Cor_BitVector& bv, Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal ihash = 0;
    HashVal arg_hash;
    int bv_i;
    for (bv_i=0; bv_i < count(); bv_i++) {
        if (bv.test(bv_i)) {
            // should probably also check that "bv" has the right count.
            //arg_hash = (*this)[bv_i]->hash(env);
            arg_hash = (*this)[bv_i].hash(env);
            if (arg_hash == VarHashValue)
                return arg_hash;
            ihash = (ihash * 31) ^ arg_hash;
        }
    }

    return Cor_IntToHash(ihash);
}

// *************************************************************************
HashVal Cor_ArgList::hash(Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal ihash = 0;
    HashVal arg_hash;
    int bv_i;
    for (bv_i=0; bv_i < count(); bv_i++) {
        //arg_hash = (*this)[bv_i]->hash(env);
      arg_hash = (*this)[bv_i].hash(env);
        if (arg_hash == VarHashValue)
            return arg_hash;
        ihash = (ihash * 31) ^ arg_hash;
    } 
    return Cor_IntToHash(ihash);
}

// *************************************************************************
// Make a true list, i.e. [a,b,c], from this arglist
Cor_CArg *Cor_ArgList::make_list()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    Cor_CArg *item;
    Cor_CArg *list = new Cor_CArg(NIL_ARG);

    for (j=count()-1; j >= 0; j--) {
        item = args[j].dup();
        list = new Cor_CArg(item,list);
    }

    return(list);
}

// *************************************************************************
// Simplify each arg with the BindEnv supplied.
int Cor_ArgList::simplify_me(Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int ret = 1;

    for (int j=count()-1; j >= 0; j--)
        ret &= args[j].simplify_me(env);

    return(ret);
}

// *************************************************************************

