// *************************************************************************
//    ArgList.h
// *************************************************************************

#ifndef _ArgList_H_
#define _ArgList_H_

#include "CArg.h"
#include "Debug.h"

// *************************************************************************

class Cor_BindEnv;
class Cor_IntStack;

// *************************************************************************

class Cor_ArgList {

    unsigned short  numArgs;    // number of arguments in ArgList

    Cor_CArg       *args;       // array of the arguments 

  public:

#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_ArgList(int size);
    Cor_ArgList(const Cor_ArgList&);
    Cor_ArgList(const Cor_CArg&);
    Cor_ArgList(Cor_CArg*, int size);    // Array already supplied
    Cor_ArgList();
   ~Cor_ArgList();

    void operator=(const Cor_ArgList&);
    void set_count(int size);
    int count() const    { return numArgs; }
    int arity() const    { return numArgs; }

    Cor_CArg& operator [] (int i) const {
        return args[i];
    }

    const Cor_CArg& arg(int i) const {
        return args[i];
    }

    // Returns true if all the arguments are Numbers or Symbols (or Arrays)
    int isConstant();

      // Returns true if the arglist has grouping args
    int has_grouping_args();


    void print_dump(Cor_BindEnv *context, ostream *file) const;
    void print(Cor_BindEnv *context, ostream *file) const;
    void print_qv(Cor_BindEnv *context, ostream *file) const;
    void print(Cor_BindEnv *context, ostream *file, char *fmt_str) const;
    void literal_print(ostream *file) const;

    friend ostream& operator<< (ostream& out, const Cor_ArgList& al);

    void printon(ostream *file) const { print((Cor_BindEnv*)0, file); }
    void sprint(char *str, int *pos, Cor_BindEnv *context = NULL) const;


    Cor_IntStack *var_list(Cor_IntStack *var_stk);
    Variable max_vars();

    HashVal hash( Cor_BitVector& bv, Cor_BindEnv *env = NULL );
    HashVal hash( Cor_BindEnv *env = NULL );

        // Make a duplicate of this ArgList
    Cor_ArgList *dup() {
          // Check for trival case
        if (!this)
            return(NULL);

        Cor_ArgList *copy = new Cor_ArgList(numArgs);
        for (int j=0; j < numArgs; j++)
           (copy->args)[j] = args[j];

        return(copy);
    }

        // Make a true list, i.e. [a,b,c], from this arglist.
    Cor_CArg *make_list();

      // Simplify each arg with the BindEnv supplied.
    int simplify_me(Cor_BindEnv *env);
};

// ; phase III. type info: ArgList ARGLIST, CArg *ARG
#define FOR_EACH_ARG(ARG, ARGLIST)    \
    { \
        Cor_CArg *ARG;  \
        int __num_for_each_arg_coral__ = (ARGLIST).count(); \
        int __i_for_each_arg_coral__ = 0;  \
        for (; (-- __num_for_each_arg_coral__) >= 0;) { \
            ARG = ( &((ARGLIST)[__i_for_each_arg_coral__++]));

#define END_EACH_ARG \
    } \
}

// *************************************************************************

#endif _ArgList_H_
