// *************************************************************************
//    BindEnv.C
// *************************************************************************

#include <string.h>

#include "BindEnv.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_BindEnv::Cor_BindEnv(int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _limit = size;
    bindings = new Cor_Term [size];
}

// *************************************************************************
Cor_BindEnv::Cor_BindEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _limit = 0;
    bindings = NULL;
}

// *************************************************************************
Cor_BindEnv::~Cor_BindEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

#ifdef    DEBUG
   if (Cor_CurTrail->is_InTrail(this)) {
     cout << "ERROR in " << __FUNCTION__ << ": BindEnv " << this
          << " is still present in the Trail.\n";
     callstk->show_call_stack(&cout);
   } 
#endif

    for (j=0; j < _limit; j++)
        if (bindings[j].bindenv != NULL) {
            cerr << "Non-null bindenv in a term in a bindenv.\n";
            ASSERT(0);
        }

    delete [] bindings;
}

// *************************************************************************
// Avoid using this method  (slow)
Cor_BindEnv *Cor_BindEnv::dup()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    Cor_BindEnv *that = new Cor_BindEnv(_limit);

    for (j=0; j < _limit; j++)
        that->bindings[j] = bindings[j];

    return(that);
}

// *************************************************************************
int Cor_BindEnv::all_free()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for(int i=0; i < _limit; i++) {
        if (bindings[i].expr.is_empty())
            return(0);
    }

    return(1);
}

// *************************************************************************
void Cor_BindEnv::resize(int newsize)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Term *newbindings = new Cor_Term [newsize];

    for (int i=0; i < _limit; i++)
       newbindings[i] = bindings[i];

    //memset(&(newbindarray[_limit]), 0, (newsize-_limit) * sizeof(Cor_Term));
    delete [] bindings;
    _limit = newsize;
    bindings = newbindings;
}

// *************************************************************************
#ifdef    NOT_USED
// This is a hack --- see note on versioning-hack for more information
// env has to be an array bindenv
void Cor_BindEnv::restore(Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_BindEnv *newenv = (Cor_BindEnv *)env;

    bindings = newenv->bindings;
    _limit = newenv->_limit;
}
#endif   // NOT_USED

// *************************************************************************

#ifdef    NOT_USED
// This is a hack -- see note on versioning-hack for more information
// Copy out the structure, but not the contents
// Use this very carefully --- deleting bindenvs will delete the
// bindarray too, so better use ::delete on the original bindenv
// rather than delete.

Cor_BindEnv *Cor_BindEnv::copy_shell()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_BindEnv *newenv = new Cor_BindEnv();
    newenv->_limit = _limit;
    newenv->bindings = bindings;

    return(newenv);
}

#endif  // NOT_USED
// *************************************************************************
void Cor_BindEnv::printon(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (this) {
        *out << "Printing BindEnv: " << this << endl;
        *out << bindings << " (size is " << _limit << ")\n";

        for (int j=0; j < _limit; j++) {

            *out << "[" << j << "]: ";
            (bindings[j]).printon(out);
            *out << endl;
        }
    }
}

// *************************************************************************
ostream& operator <<(ostream& out, const Cor_BindEnv& env)
{
    for (int j=0; j < env._limit; j++) {

        out << "[" << j << "]: ";
        out << (env.bindings)[j] << endl;
    }

    return(out);
}

// *************************************************************************
