// *************************************************************************
//    Stack.h
// *************************************************************************

#ifndef _Stack_H_
#define _Stack_H_

#include <stdlib.h>

#include "CArg.h"

// *************************************************************************

class Cor_IntStack {

  public:
    int *data;
    int  length;
    int  data_size;

    Cor_IntStack() {
                     data = NULL;
                     length = 0;
                     data_size = 0;
                   }
    Cor_IntStack(int sze) {
                            data = new int[sze];
                            length = 0;
                            data_size = sze;
                          }

   ~Cor_IntStack() { delete [] data; }

    int  count() { return(length); }
    int  is_empty() { return(length == 0); }

      // push() returns the position
    int  push(int item);

      // Warning, insert_at is not particularly efficient
    void  insert_at(int pos, int item);

    int  get_ith(int i) {
        if (i < length)
            return(data[i]);
        return(0);
    }

    void set_ith(int i, int val) {
        if (i < length)
            data[i] = val;
    }

      // Delete an element based on its position, this function must
      // maintain the "stack" concept, sliding the other values down.
    void del_ith(int i) {
        for (; i < length-1; i++) 
            data[i] = data[i+1];
        length--;
    }

        // Pull out the last element
    int pop() {
        if (length > 0) {
            length--;
            return(data[length]);
        } else {
              // Not the best, need to be careful not to exceed
            return(0);
        }
    }

        // Peek at the last element
    int top() {
        if (length > 0)
            return(data[length-1]);
        return(0);
    }

    void copy(Cor_IntStack&);

      // Empty out the stack
    void empty() {
        length = 0;
    }

    int contains(int elemnt) {
        for (int j=0; j < length; j++)
            if (data[j] == elemnt)
                return(1);
        return(0);
    }

      // Delete an element based on its value, this function must
      // maintain the "stack" concept, sliding the other values down.
      // The return value indicates if the value was found.      
    int  del_val(int val) {
        int did_it = 0;

        for (int j=0; j < length; j++)
            if (data[j] == val) {
                did_it = 1;
                for (; j < length-1; j++)
                    data[j] = data[j+1];
                length--;
                break;
            } 

        return(did_it);
    }

    int  set_val(int oldval, int newval) {
        int did_it = 0;

        for (int j=0; j < length; j++)
            if (data[j] == oldval) {
                did_it = 1;
                data[j] = newval;
            } 

        return(did_it);
    }

      // Stream printing
    friend ostream& operator<< (ostream& out, const Cor_IntStack& ilist);
};

// *************************************************************************

class Cor_VoidStack {

  public:
    void **data;
    int    length;
    int    data_size;

    Cor_VoidStack() { data = NULL; length = 0 ; data_size = 0; }
   ~Cor_VoidStack() { delete [] data; }

    int   count() { return(length); }
    int   is_empty() { return(length == 0); }

      // push() returns the position
    int   push(void *item);

      // Warning, insert_at is not particularly efficient
    void  insert_at(int pos, void *item);

    void  *get_ith(int i) {
        if (length > i)
            return(data[i]);
        return(NULL);
    }

        // Pull out the last element
    void *pop() {
        if (length > 0) {
            length--;
            void *tmp = data[length];

              // This is just so no one else try's grabbing this one 
            data[length] = NULL;
            return(tmp);
        } else {
            return(NULL);
        }
    }

    void  copy(Cor_VoidStack&);

      // Empty out the stack
    void empty() {
        length = 0;
    }
};

// *************************************************************************

class Cor_CArgStack {

  private:
    int        delete_cargs;
  
  public:
    Cor_CArg **data;
    int        length;
    int        data_size;
    int        ref_cnt;    // For reference counting this item

    Cor_CArgStack(int dp=1) {
        data = NULL;
        length = 0;
        data_size = 0;
        delete_cargs = dp;
        ref_cnt = 0;
    }

   ~Cor_CArgStack() {
        if (delete_cargs)
            empty();
        delete [] data;
    }

      // Stream printing
    friend ostream& operator<< (ostream& out, const Cor_CArgStack& ca_list);

    int  count() { return(length); }
    int  is_empty() { return(length == 0); }

      // push() returns the position
    int  push(Cor_CArg*);

      // Warning, insert_at is not particularly efficient
    void  insert_at(int pos, Cor_CArg *item);

    void clr_ith(int i) {
        if (i < length)
            data[i] = NULL;
    }

    Cor_CArg *get_ith(int i) {
        if (length > i)
            return(data[i]);
        return(NULL);
    }

    Cor_CArg *dup_ith(int i) {
        if (length > i)
            return(data[i]->dup());
        return(NULL);
    }

        // Pull out the last element
    Cor_CArg *pop() {
        if (length > 0) {
            length--;
            Cor_CArg *tmp = data[length];

              // This is just so no one else try's grabbing this one
            data[length] = NULL;
            return(tmp);
        } else {
            return(NULL);
        }
    }

    void copy(Cor_CArgStack&);
    void copy_bottom(Cor_CArgStack& stk, int num_items);
    void build_arglist(Cor_ArgList& args);

      // Empty out the stack - delete each element
    void empty();

      // Duplicate the Stack (and all arguments)
    Cor_CArgStack *dup() {
        Cor_CArgStack *ca_list = new Cor_CArgStack();
        ca_list->copy(*this);
        return(ca_list);
    }

      // Duplicate the Stack as far as specified (used when storing lists)
    Cor_CArgStack *dup_bottom(int num_items) {
        Cor_CArgStack *ca_list = new Cor_CArgStack();
        ca_list->copy_bottom(*this,num_items);
        return(ca_list);
    }
};

// *************************************************************************

#endif _Stack_H_
