// *************************************************************************
//    ParseActions.C
// *************************************************************************

#include "ParseActions.h"
#include "ParserStruct.h"
#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "Globals.h"
#include "Execute.h"
#include "Literal.h"
#include "Database.h"
#include "Interface.h"
#include "Tuple.h"
#include "BindEnv.h"
#include "Rewriting.h"
#include "Compile.h"
#include "Unparse.h"
#include "Debug.h"

// *************************************************************************

#define TMP_RELNAME  "__RelTemp"

// *************************************************************************

static int is_recursive(Cor_ParserStruct *Cor_parserStruct);

// *************************************************************************
void Cor_AppendFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark;

    if (Cor_exEnv->in_query_loop) {

        if (is_recursive(Cor_parserStruct)) {
            *(Cor_exEnv->err_strm) << "CORAL::error :";
            *(Cor_exEnv->err_strm) <<
                "Recursion not allowed in imperative rule!\n";
        } else {
            Cor_complete_rule(Cor_parserStruct->cur_rule, Cor_parserStruct);
            int num_ans = Cor_execute_single_rule(Cor_parserStruct);
            if (Cor_exEnv->C_verbose_answer_default) {
                *(Cor_exEnv->out_strm) << 
                    "(Number of Tuples to Append = " << num_ans << ")\n";
            }
        }

    } else {
        *(Cor_exEnv->err_strm) << "CORAL::error :";
        *(Cor_exEnv->err_strm) << "Imperative statement inside module!\n";
    }

    if (Cor_exEnv->C_verbose_answer_default)
        Cor_exEnv->out_strm->flush();
    stack_mark.pop_to();   
}

// *************************************************************************
void Cor_DeleteFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark;

    if (Cor_exEnv->in_query_loop) {

        if (is_recursive(Cor_parserStruct)) {
            *(Cor_exEnv->err_strm) << "CORAL::Error :"
                << "Recursion not allowed in an imperative rule!\n";
        } else {
            Cor_complete_rule(Cor_parserStruct->cur_rule, Cor_parserStruct);

            Cor_CArg headName = Cor_parserStruct->cur_rule->head->pred;
            Cor_CArg tmpName  = Cor_CArg(TMP_RELNAME);

            // rename the <head relation> to TMP_NAME
            if (!Cor_renameRelation(headName, tmpName)) {
                *(Cor_exEnv->err_strm) << "unable to rename file\n";
                return;
            }

            // the tuples to delete are inserted into a new <head relation>
            int num_ans = Cor_execute_single_rule(Cor_parserStruct);

            Cor_Relation *rel = Cor_find_external_relation(headName,
                    Cor_parserStruct->cur_rule->head->arity());
            
            Cor_Relation *temprel = Cor_find_external_relation(tmpName,
                    Cor_parserStruct->cur_rule->head->arity());

            if (!rel || !temprel) {
                *(Cor_exEnv->err_strm) << "Strange brew !!\n";
                return;
            }

            // Then delete the tuples in <head relation> from TMP_NAME
            Cor_Tuple *tuple ;
            FOR_EACH_TUPLE(tuple, rel) {
                temprel->tuple_delete(tuple);
            } END_EACH_TUPLE(tuple);

            // Finally, close the <head relation>
            rel->empty_relation(0);
            Cor_CurDB->RelationTable->delete_entry(headName);
            delete rel;

            // rename TMP_NAME to <head relation>
            if (!Cor_renameRelation(tmpName, headName)) {
                *(Cor_exEnv->err_strm) << "unable to rename tmp file\n";
                return;
            }

            if (Cor_exEnv->C_verbose_answer_default) {
                *(Cor_exEnv->out_strm) << 
                    "(Number of Tuples to Delete = " << num_ans << ")\n";
            }
         }

    } else {
        *(Cor_exEnv->err_strm) << "CORAL::error :"
            << "Imperative statement inside module!\n";
    }

    if (Cor_exEnv->C_verbose_answer_default) 
        Cor_exEnv->out_strm->flush();

    stack_mark.pop_to();
}

// *************************************************************************
void Cor_AssignFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark;

    if (Cor_exEnv->in_query_loop) {
        if (is_recursive(Cor_parserStruct)) {
            *(Cor_exEnv->err_strm) << "CORAL::error :";
            *(Cor_exEnv->err_strm) << 
                "Recursion not allowed in imperative rule!\n";
        } else {
            Cor_complete_rule(Cor_parserStruct->cur_rule, Cor_parserStruct);

            Cor_Relation *rel = NULL;

            if (Cor_parserStruct->cur_rule) {
                rel = Cor_find_external_relation(
                    Cor_parserStruct->cur_rule->head->pred,
                    Cor_parserStruct->cur_rule->head->arity());
            }

            // Clear the lhs relation.
            if (rel) {
                rel->empty_relation(0);
            }

            // now execute the rule
            int num_ans = Cor_execute_single_rule(Cor_parserStruct);

            if (Cor_exEnv->C_verbose_answer_default) {
                *(Cor_exEnv->out_strm) << 
                    "(Number of Tuples to Assign = " << num_ans << ")\n";
            }
        }

    } else {
        *(Cor_exEnv->err_strm) << "CORAL::error :";
        *(Cor_exEnv->err_strm) << "Imperative statement inside module!\n";
    }

    if (Cor_exEnv->C_verbose_answer_default) 
        Cor_exEnv->out_strm->flush();

    stack_mark.pop_to();   
}

// *************************************************************************
void Cor_QueryFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark;
    Cor_CArg *varSymbol;
    Cor_CArg  headSymbol;

    if (!Cor_exEnv->in_query_loop) {
        *(Cor_exEnv->err_strm) << "CORAL::error :";
        *(Cor_exEnv->err_strm) << "Imperative statement inside module!\n";
        return;
    }
    
#   ifdef DO_PROFILE
    Cor_GlobalProfileInfo.clear();
#   endif

    Cor_Clause *cur_rule = Cor_parserStruct->cur_rule;

    if (Cor_exEnv->C_answer_style_default == COR_EMPTY_ANSWER_PRINT)
        headSymbol = *Cor_PrintEmptySymbol;
    else if (Cor_exEnv->C_answer_style_default == COR_TUPLE_ANSWER_PRINT)
        headSymbol = *Cor_PrintTupleSymbol;
    else if (Cor_exEnv->C_answer_style_default == COR_LITERAL_ANSWER_PRINT)
        headSymbol = *Cor_PrintLiteralSymbol;
    else
        headSymbol = *Cor_PrintBindingSymbol;

    // make print() the head of the rule 
    cur_rule->head = new Cor_Literal(Cor_parserStruct, headSymbol,
        Cor_parserStruct->rule_var_names.length);

    // create the arglist of print() to have all the
    // variables in the body of the query 

    Cor_CArg var_carg;

    for (int i = 0; i < cur_rule->head->arity(); i++) {

        varSymbol = Cor_parserStruct->rule_var_names.data[i];

        // short-circuited this for the common case - SPF
        if (varSymbol->equals(Cor_DontCareSymbol)) {
            varSymbol = varSymbol->dup();
            int vn = Cor_varname_enter(&(Cor_parserStruct->rule_var_names),
                          varSymbol);
            var_carg.set_var(vn);
            cur_rule->head->args[i] = var_carg;
        } else {
            var_carg.set_var(i);
            cur_rule->head->args[i] = var_carg;
        }
    }

    Cor_exEnv->C_interactive_mode_current =
            Cor_exEnv->C_interactive_mode_default;

    Cor_complete_rule(Cor_parserStruct->cur_rule, Cor_parserStruct);

    int num_ans = Cor_execute_single_rule(Cor_parserStruct);

    if (Cor_exEnv->C_verbose_answer_default) {
        *(Cor_exEnv->out_strm) << "(Number of Answers = "
            << num_ans << ")\n" << flush;
    }

    stack_mark.pop_to();
}

// *************************************************************************
void Cor_FactFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_parserStruct->rhspreds.length > 0) {
        *(Cor_exEnv->err_strm) << "CORAL::error : "
             << "Rule declared outside a module!\n";
    } else {

      Cor_Relation *rel;
      rel = Cor_find_external_relation(Cor_parserStruct->cur_rule->head->pred,
              Cor_parserStruct->cur_rule->head->arity());

      // use the kludge that ensures that all builtins have count = -1
        
      if (!(Cor_exEnv->C_insert_mode_default) && !Cor_parseStack->is_empty() &&
          (Cor_parseStack->top_file() == Cor_curConn->conn_strm))
      {
          if (!rel || (rel->count >= 0)) {
              *(Cor_exEnv->err_strm) << "CORAL::error : "
                                     << "Not in insert mode !\n";
              return;
          }
      }
      
      if (!rel) {
          rel = Cor_make_relation(Cor_parserStruct->cur_rule->head->pred,
              Cor_parserStruct->cur_rule->head->arity());
      }
  
      rel->insert_new(Cor_parserStruct->cur_rule->head->args,
                (Cor_BindEnv*)NULL);
    }
}

// *************************************************************************
void Cor_EndModuleFunction(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_parserStruct->CurModule->PreProcessing)
        Cor_perform_rewriting(Cor_parserStruct);
    else
        Cor_interpret_module(Cor_parserStruct);
}

// *************************************************************************
static int is_recursive(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *head = Cor_parserStruct->cur_rule->head;
    Cor_Literal *body_pred;

    for (int i = 0; i < Cor_parserStruct->rhspreds.length; i++) {
        body_pred = (Cor_Literal*)Cor_parserStruct->rhspreds.data[i] ;
        if ((head->pred.equals(body_pred->pred)) &&
                (head->arity() == body_pred->arity()))
        {
            return(1);
        }
    }
    return(0);
}

// *************************************************************************
