// *************************************************************************
//    ParserEnv.C
// *************************************************************************

#include "ParserEnv.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Debug.h"
#include "WodiStream.h"

extern "C"     int yylex(YYSTYPE *yylval);

// *************************************************************************
void Cor_ParserEnv::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    cur_line_num = 0;
    prompt = coralConfig->prompt;
    strcpy(prompt_name,prompt);
    prompt = prompt_name;
    show_prompt = Cor_exEnv->C_print_prompt_default;

    history = new Cor_History(MAX_HISTORY_DEPTH);
    command_count = 0;
    is_history_command = 0;

    linebuf[0] = '\0';
    line_offset = 0;
    commandbuf[0] = '\0';
    command_offset = 0;

    in_comment = 0;

    in_strm = (CoralStream *) NULL;

    C_num_head_arith_lits = 0;
    complexRelation = 0;

        // This line is NOT necessary, as lexer is always set,
        // but its better to be safe then sorry. 
    lexer = (LEXER) yylex;

    next = (Cor_ParserEnv *) NULL;
}

// *************************************************************************
Cor_ParserEnv::~Cor_ParserEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete history;
}

#ifdef DO_WODI
extern to_wodi_flush;
#endif //DO_WODI

// *************************************************************************
// This is a hack to avoid the select problem that keeps coming up, the
// client should reply with another command, or a newline if no command
// is ready.
void Cor_ParserEnv::request_ack()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_WODI
#else

    if (coralConfig->server_mode){
          // Modified to use Client protocol
        *(Cor_exEnv->out_strm) << CP_ACK << endl << flush;
    }

#endif // DO_WODI
}

// *************************************************************************
void Cor_ParserEnv::print_prompt()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_WODI
    if (coralConfig->server_mode) {
        *(Cor_exEnv->out_strm) << CP_OKAY << endl << flush;

        if (to_wodi_flush) {
            ((Cor_wodibuf*)Cor_exEnv->out_strm->rdbuf())->wodi_flush();
            to_wodi_flush = 0;
        }
    }
    else if (show_prompt && Cor_exEnv->C_print_prompt_default) {
        *(Cor_exEnv->out_strm) << prompt << flush;
    }

#else

    if (show_prompt){        
        if (coralConfig->server_mode){
              // Modified to use Client protocol
            *(Cor_exEnv->out_strm) << CP_OKAY << endl << flush;

        } else if (Cor_exEnv->C_print_prompt_default) {
            *(Cor_exEnv->out_strm) << prompt << flush;
        }
    }
#endif // DO_WODI
}

// *************************************************************************
void Cor_ParserEnv::handle_history()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (incr_command_count && !is_history_command)
       (history)->addCommand(commandbuf, ++command_count);
     
    if (line_offset + command_offset >= MAX_RULE_LENGTH) {
       *(Cor_exEnv->err_strm) << 
           "CORAL::Warning : Line buffer size exceeded!\n";
    }

    strncat(linebuf, commandbuf, command_offset);
    line_offset += command_offset;
    commandbuf[0] = '\0';
    command_offset = 0;
    is_history_command = 0;
}

// *************************************************************************
void Cor_ParserEnv::fix_history()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    commandbuf[0] = '\0';
    command_offset = 0;

    linebuf[0] = '\0';
    line_offset = 0;
}

// *************************************************************************
