// *************************************************************************
//    ContextNode.h
// *************************************************************************

#ifndef _ContextNode_H_
#define _ContextNode_H_

#include "Context.h"
#include "GoalNode.h"
#include "GoalNodeLink.h"
#include "Debug.h"

// *************************************************************************
//    Codependent classes

class Cor_Context;
class Cor_ContextNode;

// *************************************************************************

class Cor_ContextNode {

  public:
    Cor_GoalNode    *goals;

    Cor_ContextNode *all_next;        // For memory cleanup -SPF
    Cor_ContextNode *all_prev;

    // unmarked goals as a singly linked list
    Cor_GoalNodeLink *first_unmarked_goal;    
    Cor_GoalNodeLink *last_unmarked_goal;

    // marked goals as a singly linked list
    Cor_GoalNodeLink *first_marked_goal;
    Cor_GoalNodeLink *last_marked_goal;

    int mark;            // if any goal is marked, so is the node
    int node_number;     // used for linear ordering of nodes

    int negated;    
    // CHECK IF NEEDED??
    // if any goal is negated, so is the node

    Cor_ContextNode *next;    // link to next node in context
    Cor_ContextNode *prev;    // link to previous node in context
    Cor_ContextNode *next_marked;    // link to next marked node
    Cor_ContextNode *prev_marked;    // link to previous marked node

    Cor_Context *context;    // the context to which this belongs to 

    Cor_ContextNode(Cor_GoalNode *new_goal, Cor_Context *c);
   ~Cor_ContextNode();
    
    inline int is_marked()        { return mark; }
    inline void mark_node()        { mark = 1; }

    int any_unmarked_goals();
    Cor_GoalNode *get_unmarked_goal();

    void move_goals_to_done();

    void move_to_marked_list(Cor_GoalNodeLink *g_ptr);
    void merge_nodes(Cor_ContextNode *tbm);

    void update_goalnode_info();
    void print_all(ostream *file);
    void free_all();
};

// *************************************************************************

#endif _ContextNode_H_
