// *************************************************************************
//    SemiNaive.C
// *************************************************************************

#include <stdio.h>

#include "SemiNaive.h"

#include "SccInfo.h"
#include "PredAnnotations.h"
#include "CompileRule.h"
#include "Compile.h"
#include "Relation.h"
#include "Unparse.h"
#include "Globals.h"
#include "Debug.h"


// *************************************************************************
Cor_SemiNaive::Cor_SemiNaive()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    scc = NULL;
    rInfo = NULL;
    first = 0;
    markOffsets = NULL;
    negated = NULL;
    predOrders = NULL;
    firstFailureBacktrack = NULL;
    nextFailureBacktrack = NULL;  
    unify_funct = NULL;
    successBacktrack = 0;
    recursive = 0;
    indices = NULL;
    ignore_this_rule = 0;
    rule_id = -1;
    label = NULL;
}

// *************************************************************************
Cor_SemiNaive::~Cor_SemiNaive()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    if (rInfo) {
        delete markOffsets;
        delete [] indices;
        delete [] negated;
        delete [] predOrders;
        if (rInfo->num_literals) {
            delete [] firstFailureBacktrack;
            delete [] nextFailureBacktrack;
        }

        if (unify_funct)
            for (j=0; j < rInfo->num_literals; j++)
                delete [] unify_funct[j];
        delete [] unify_funct;
    }
    delete [] label;

    // SPF - This is a bad idea, its recursive
    //delete scc;
}

// *************************************************************************
void Cor_SemiNaive::clean_cache() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i=0; i < rInfo->num_literals; i++)
        indices[i] = -1;
}

// *************************************************************************
void Cor_SemiNaive::Add_Rule_Info(Cor_RuleInfo *rule_info,
                  Cor_Clause * /* clause */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    rInfo = rule_info;

    delete [] markOffsets;
    markOffsets = new int[(rInfo->num_literals)*2];
#ifdef    DEBUG
    memset(markOffsets,0, sizeof(int)*(rInfo->num_literals)*2);
#endif

    delete [] indices;
    indices = new int [rInfo->num_literals];

    for (i = 0; i < rInfo->num_literals; i++)
        indices[i] = -1;

    unify_funct = new Cor_Unify_Funct* [rInfo->num_literals];

    for (i = 0; i < rInfo->num_literals; i++)
        unify_funct[i] = NULL;
    

    predOrders = new int[rInfo->num_literals];
    Cor_initPredOrders(this);
        
    // Generate index annotations if the SemiNaive rule is a part of an scc.
    // In this case, the scc belongs to a module, and index annotations
    // can be created for the module. However, if the rule is a single
    // imperative rule (executed from the CORAL prompt), there is no
    // module for it. A separate index annotation list is created and
    // used to store the index annotations. These annotations are executed
    // right away resulting in eager index creation.
    // NOTE:: generateIndexes can change the PredOrders array ! Hence
    // anything that uses the predOrders should come after it.      

    if (scc)
        Cor_generateIndexes(this, scc->module->index_annotations);
    else {
        Cor_PredAnnotations *tmp_anno = NULL, *cur;
        Cor_generateIndexes(this, tmp_anno);

        for(cur= tmp_anno; cur;) {
            int mx1, mx2 = 0;
            mx1 = cur->arglist1->max_vars();
            if ( cur->arglist2 )
                mx2 = cur->arglist2->max_vars();
            int num_vars = (mx1 > mx2) ? mx1 : mx2;

            Cor_Relation *indexrel = Cor_find_external_relation(cur->pred,
                cur->arglist1->arity());

            if (indexrel) {
                if (cur->arglist2)
                    indexrel->add_index(cur->arglist1,num_vars,cur->arglist2);
                else
                    indexrel->add_index(cur->arglist1);
            }
            // delete space for index annos
            Cor_PredAnnotations *temp = cur;
            cur = cur->next();
            delete temp;
        }    
    }

    // NOTE: predOrders MUST be assigned BEFORE the backtrack arrays
    // and negated arrays

    delete [] negated;
    negated = new int[rInfo->num_literals];
    for (i = 0; i < rInfo->num_literals; i++)
        negated[i] = (rInfo->relations[i] < 0);

    if (rInfo->num_literals) {

            // In case these were already in use?
        delete [] firstFailureBacktrack;
        delete [] nextFailureBacktrack;

        firstFailureBacktrack = new int [rInfo->num_literals];
        nextFailureBacktrack = new int [rInfo->num_literals];
        successBacktrack = initFailureBacktrack(this);
        btDecision[BACKWARD] = nextFailureBacktrack;
        btDecision[FORWARD] = firstFailureBacktrack;

        if (scc && scc->module->module_info->NaiveBacktrack) {
            for (int j=0; j < rInfo->num_literals; j++) 
                firstFailureBacktrack[j] = nextFailureBacktrack[j];
        }
    }
    
    clean_cache();

    // this is an optimization for supplementary magic with indexing
    // the magic rules are ignored, and the magic facts are generated
    // along withthe firing of the goalid() literal of the supp rule
    //              --- Praveen
    // note that scc could be NULL if this is an imperative rule invoked
    // from the coral prompt

    /******* This causes problems with return unify optimization : Praveen
    if (scc && (scc->module->module_info->SupMagicIndexing) && 
          is_magic(clause->head->pred))
        ignore_this_rule = 1;
    *****/
}

// *************************************************************************
void Cor_SemiNaive::print(ostream *fd)
{ // print rule info
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  //  rInfo->pInfo.print(fd, "", predOrders, rInfo->num_literals);
  *fd << "  sn_rule #" <<  rule_id << ": ";
  *fd <<  label << ": ";
  Cor_print_rule_with_order(rInfo->rule,predOrders, markOffsets, recursive,fd);
    
}

// *************************************************************************
void Cor_SemiNaive::Add_label()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  label = new char[20];

  int org_rule_no, magic_rule_no;
  int j;
  
  // get the original rule number
  j = rInfo->rule->clause_names->count();
  if ((j == 0)
       || (sscanf(rInfo->rule->clause_names->ith_symbol(0),
               "%d", &org_rule_no) < 1))
      org_rule_no = 0;

  // get the magic rule number, if it has one
  if ((j <= 1)
       || (sscanf(rInfo->rule->clause_names->ith_symbol(1),
               "%d", &magic_rule_no) < 1))
        magic_rule_no = 0;

  // form a label for the SN rule using the original rule no, magic rule no
  // and the label counter
  sprintf(label, "%d.%d.%d", org_rule_no, magic_rule_no, 
      rInfo->rule->sn_rule_label_counter);

  rInfo->rule->sn_rule_label_counter++;
}

// *************************************************************************
